Mojo README
===========

Mojo is a deployment workflow system for Juju.

Mojo is a continuous integration and continuous delivery system for Juju-based
services managed by Canonical IS. It is intended to replace the current system
of scripts and manual manipulation used to deploy and maintain services.

At a high level, Mojo is responsible for testing deployment of new services
upgrades to existing services. Mojo monitors a deployment specification for
changes and automatically test-deploys based on that spec. New specs are
deployed against existing environments, but from-scratch deployments can also
be tested regularly to ensure that services could be rebuilt.

The power of mojo is versatility. By creatively organizing phases and
configuration files many deployment permutations can be tested.

Usage
-----

Mojo provides several commands, each performing one phase of a Juju deployment
process. These commands reference a Mojo spec for details about the
environment, charms, and deployment processes.

In addition to commands that control your deployment, Mojo includes tools for
working with specs, preparing your deployment, and for integrating with CI
systems like Jenkins.

### Projects ###

A mojo project is the main staging ground for testing. It is directory where 
deployment artifacts are assembled. A mojo project also has an lxc container
associated with it for the build process.

Typically there is one project per juju environment. There can be many 
workspaces per project.

The root MOJO_DIR is typically /srv/mojo so projects are located:

    /srv/mojo/$MOJO_PROJECT

The LOCAL directory is where secrets and certificates or anything that cannot be
stored in the spec repository should be staged for use by the deployment.
The mojo secrets phase copies data from the LOCAL directory to the workspace
local directory.

    /srv/mojo/LOCAL

The ROOTFS directory is the lxc container rootfs.

    /srv/mojo/$MOJO_PROJECT/$MOJO_SERIES/ROOTFS

with $MOJO_SERIES as either precise, trusty, etc.

### Workspaces ###

A mojo workspace is staging ground for the specific version of the specification
being tested. Typically the workspace is named for the revision number of the 
spec and the stage. i.e 23-production

    /srv/mojo/$MOJO_PROJECT/$MOJO_SERIES/$MOJO_WORKSPACE

Workspace layout:

The specification is pulled from a bzr repository to the spec directory.

    /srv/mojo/$MOJO_PROJECT/$MOJO_SERIES/$MOJO_WORKSPACE/spec

The build directory is where collect phase stores the charms and payload data
by default. The build phase script should point to the contents in this
directory.

    /srv/mojo/$MOJO_PROJECT/$MOJO_SERIES/$MOJO_WORKSPACE/build

The charms directory is where the repo phase setups up the charm repository.
The deploy phase deploys charms from this location.

    /srv/mojo/$MOJO_PROJECT/$MOJO_SERIES/$MOJO_WORKSPACE/charms

The local directory is where secrets and certificates or anything that cannot be
stored in the spec repository should be moved to for use by the deployment.
The mojo secrets phase copies data from the LOCAL directory to the workspace
local directory.

    /srv/mojo/$MOJO_PROJECT/$MOJO_SERIES/$MOJO_WORKSPACE/local


Setup Commands
--------------

Create a new project:

    sudo mojo project-new $MOJO_PROJECT -s $MOJO_SERIES
    # eg:
    sudo mojo project-new my-stack-name -s precise


List all projects

    mojo project-list

Delete a project

    sudo mojo project-destroy $MOJO_PROJECT

Create a new workspace:
    
    mojo workspace-new --project $MOJO_PROJECT -s $MOJO_SERIES $SPEC_URL $MOJO_WORKSPACE
    # eg:
    mojo workspace-new --project my-stack-name -s precise lp:~user/mojo-specs/my-stack-name stack-test-wip

List all workspaces

    mojo workspace-list --project $MOJO_PROJECT -s $MOJO_SERIES
    # eg:
    mojo workspace-list --project my-stack-name -s precise

Delete a workspace

    mojo workspace-destroy --project $MOJO_PROJECT -s $MOJO_SERIES $MOJO_WORKSPACE
    # eg:
    mojo workspace-destroy --project my-stack-name -s precise stack-test-wip

Phases
------

Mojo arranges deployments into several phases in order to ensure their
repeatability. By breaking a deployment into small parts, Mojo allows you to
easily build separate staging and production (or any other) environments whose
differences are carefully controlled. You can also gather important information
about each phase of your deployment to help you improve the quality and
stability of your service.

Each phase may be performed any number of times, in any order. The typical
progression for a Mojo deployment is: collect, builddeps, build, repo, services,
relations, verify.

All phases require the --project, --series, --workspace and --stage switches.
These values can also be provided by the environment variables MOJO_PROJECT,
$MOJO_SERIES, $MOJO_WORKSPACE and $MOJO_STAGE


Phase Commands
--------------

### mojo run ###

The one command to rule them all. The mojo run command pulls down the spec url, 
reads the manifest and begins executing each phase in order per the manifest.

    mojo run --project $MOJO_PROJECT --series $MOJO_SERIES --stage $MOJO_STAGE $SPEC_URL $MOJO_WORKSPACE
    # eg:
    mojo run --project my-stack-name --series precise --stage devel lp:~user/mojo-specs/my-stack-name stack-test-wip


### mojo secrets ###

The secrets phase copies secrets files from a staging directory (Default:
/srv/mojo/LOCAL/$MOJO_PROJECT/$MOJO_STAGE) to the workspace local directory
(/srv/mojo/LOCAL/$MOJO_PROJECT/$MOJO_STAGE) where they can be accessed by the
remaining phases.

> Options: 
>   config=STAGED_SECRETS_DIRECTORY
>     Directory where secrets are staged
>     Default: /srv/mojo/LOCAL/$MOJO_PROJECT/$MOJO_STAGE


### mojo collect ###

The collect phase gathers all code and other resources required to build your
service via codetree. Some services may be composed only of charms, while others
may collect static content, local configuration, or code that must be compiled.

The collect configuration file is a valid codetree configuration file. The
format is the following where destination is relative to
$MOJO_WORKSPACE_DIR/build and source is a remote resource or a local directory.
Example:

> destination             source1
> ../spec/destination     source2

> Options: 
>   config=COLLECT_CONFIG_FILENAME
>     The name of the codetree collect configuration file


### mojo builddeps ###

The builddeps phase prepares the lxc container by installing any required
packages, apt repositories or apt repository gpg keys. It is passed the package,
repos or repo_keys options using a comma delimited list.

> Options: 
>   repo_keys=APT_GPG_KEY[,APT_GPG_KEY]
>     Comma delimited list of apt repository GPG key file names
>   repos=REPO[,REPO]
>     Comma delimited list of apt repositories (See man apt-add-repository)
>   packages=PACKAGE_NAME[,PACKAGE_NAME]
>     Comma delimited list of packages.


### mojo script ###

A script phase simply executes a user-defined script. Use scripts to manipulate
your Juju units directly or to set up external resources.

The script phase can be used for testing juju upgrade-charm $CHARM after an
initial deploy has completed.

A script phase run outside the lxc container may be needed to access 
nova or juju commands without the need to copy nova credentials into
the lxc container.

Environment variables are injected into the script environment (Build and verify
are subclasses of script and get these too).

    MOJO_PROJECT = Mojo project name
    MOJO_WORKSPACE = Workspace name
    MOJO_WORKSPACE_DIR = Workspace path
    MOJO_BUILD_DIR = Workspace build dir
    MOJO_REPO_DIR = Workspace repo dir
    MOJO_SPEC_DIR = Workspace spec dir
    MOJO_LOG_DIR = Workspace log dir
    MOJO_LOCAL_DIR = Workspace local dir
    MOJO_STAGE = Stage 

> Options: 
>   config=SCRIPT_NAME 
>     The name of the script to be executed
>   lxc=True/False
>     Whether to execute the script inside the lxc container
>     Defaults to False, except for the build phase which forces
>     lxc=True.
>   KEY=VALUE
>     All KEY=VALUE paramaters will be set as environment variables for use
>     by the script

### mojo build ###

A build phase is nearly identical to a script phase, except that network access
in the lxc container is prohibited during execution. This ensures that all 
required resources are documented and available during collect.

Options: 
  config=BUILD_SCRIPT_NAME 
    The name of the build script to be executed


mojo repo
---------
The repo phase assembles a local Juju charm repository from a charm-to-directory
mapping.

By building a separate repo (rather than using the build directory), Mojo allows
the developer to structure the build directory in any way s/he likes. The charm
map also becomes an explicit list of the charms expected to be used in a
deployment.

The repo configuration file is a text file of space-delimited key-value pairs,
one per line. Each key is the name of a charm. Each value is the path, relative
to $MOJO_SPEC_DIR/build, where the charm is located. If the charm name and path
are the same, the path may be omitted.

Options: 
  config=REPO_CONFIG_FILENAME 
    The name of the repo configuration file


mojo deploy
-----------
During a deploy phase, charms are deployed and/or related via juju-deployer. 

The deploy configuration file is a valid juju-deployer configuration file.

The deploy phase is often separated into services and relations which gives the
tester a better idea where failures are occurring.

Options: 
  config=DEPLOY_CONFIG_FILENAME 
    The name of the deploy juju-deployer configuration file.
  local=SECRETS_FILENAME
    The name of the local juju-deployer configuration file. Found in the
    workspace local directory. It contains secrets or other information that
    cannot be included in the spec repository.
  juju=JUJU_ENV
    The juju environment to deploy in from environments.yaml.
  delay=DELAY_IN_SECS
    The juju deployer delay between instance deploys in seconds. Default: 30
    seconds


mojo volumes
------------
The volumes phase mounts any persistent storage required by the stack.

The volumes configuration file contains a service name and nova volume ids for
the number of units for the service space delimited:
service novaid novaid

  config=VOLUMES_CONFIG_FILENAME 
    The name of the volumes configuration file.


mojo sleep
------------
Inject wait time between phases. Used when a juju command such as upgrade-charm
has been executed and time is needed before the next phase begins.

  config=SECONDS_TO_SLEEP
    Number of seconds to sleep


mojo verify
-----------
The verify phase is also a particular type of script phase. During verify, tests
are performed to ensure that the previous phases have resulted in a working
service. When a verify step succeeds, the deployment is considered as success.

Options:
  config=VERIFY_SCRIPT_NAME 
    The name of the verify script to be executed
  lxc=True/False
    Whether to execute the script inside the lxc container


Specs
=====
A Mojo spec is simply a repository of files. Currently only Bazaar repositories
are supported. The spec includes a manifest file which is used by mojo run to
run all configured phases.

Mojo is to juju as specs are to charms. The specs are where the real power of
mojo comes to light. By creatively arranging the phases and corresponding
configurations and scripts one can test almost anything. Initial deployments, 
code updates, charm updates, and order of relations all can be tested using
various combinations of phases and configuration.

spec layout
-----------
There are two supported approaches to how you structure your spec repository.

1. Single spec in one repository

At the top of the directory is the manifest which lists which phases are
executed in which order. 

Then there are the stage directories. Stages can be production, staging, devel
or any other stage you wish to keep organized and distinct.

Inside the stage directory are the configuration files and scripts which
correspond to the phases listed in the manifest. Mojo will check the $MOJO_STAGE
directory recursively.

Example spec layout:
-----------------

./manifest
./devel/collect 
./devel/build
./devel/services
./devel/relations
./devel/verify
./shared/helper/build
./shared/helper/collect

2. Multiple specs in one repository

This is best illustrated with an example file layout.

Example spec layout:
-----------------

./team1
./team1/service1
./team1/service1/manifest
./team1/service1/devel/collect
./team1/service1/devel/build
./team1/service1/devel/services
./team1/service1/devel/relations
./team1/service1/devel/verify
./team2
./team2/service2
./team2/service2/manifest
./team2/service2/devel/collect
./team2/service2/devel/build
./team2/service2/devel/services
./team2/service2/devel/relations
./team2/service2/devel/verify

In this example, you have two specs within the same repository. The first would
be called with a "stage" of "team1/service1/devel" and the second would be
called with a "stage" of "team2/service2/devel".

The nesting of directories can be arbitrary based on your needs. If you only
need one level of nesting you could have "service1/devel" or if you need three
levels of nesting you could have "department1/team1/service1/devel".

The spec manifest
-----------------
A manifest defines the deployment progression. It lists phases, one per line,
along with optional key=value options pointing to scripts or configuration
files in the stage directory.

A simple example:

    secrets
    collect
    build
    deploy config=services
    deploy config=relations
    verify

A more interesting example including an upgrade charm script:

    builddeps repo_keys=ubuntu-archive-keyring.gpg repos=deb http://ubuntu-cloud.archive.canonical.com/ubuntu precise-updates/cloud-tools main" packages=bzr,make,python-virtualenv,libpq-dev,python-dev,zip
    secrets
    collect
    build
    repo
    script config=predeploy lxc=True MYVAR=True
    # Deploy services requiring persistent storage
    deploy config=services-with-storage
    volumes config=volumeconfig
    # Deploy main services
    deploy config=services local=secrets
    # Relate subordinate charms
    deploy config=subordinate-relations
    # Relate charms
    deploy config=relations
    script config=postdeploy MYVAR=False
    verify
    # Rebuild with updated charm or payload
    collect config=collect-upgrade
    build
    repo
    # Upgrade charms
    script config=upgrade-charm
    verify

Any comment in a manifest (line beginning with a single '#') will be included
in the logged output when the spec is run. To avoid this behaviour you can
also add a comment with a double-hash ('##') which will not be included
in the logged output.

Note that we also support includes. So, our manifest might look like this:

collect
repo
deploy config=services
include config=other-manifest

We'd have a file called "other-manifest" in the same directory as "manifest"
and it might contain this:

script config=post-deploy
verify

"mojo run" would do the equivalent of running a manifest like this:

collect
repo
deploy config=services
script config=post-deploy
verify

But we could also run "mojo run --manifest-file other-manifest" that would only do:

script config=post-deploy
verify

Secrets
=======

Secrets handling is of significant importance. Secrets play a major role in the
complexity of deploying a production environment and therefore need to be tested
along with the rest of the stack.

Although the goal is to organize as much as possible in the specification
repository, secrets must be kept separate from the repository for security.
Therefore, there is a manual element to handling secrets. They must be staged
for use by mojo.

Mojo will assume secrets are staged at this location:
 /srv/mojo/LOCAL/$MOJO_PROJECT/$MOJO_STAGE

The mojo secrets phase will copy them from the staged location to the workspace
local directory:
 /srv/mojo/$MOJO_PROJECT/$MOJO_WORKSPACE/local

Add the secrets phase to the manifest early enough to make secrets available to
the rest of the phases
 This makes the secrets available inside the LXC container
 They are then available to the script phases as $MOJO_LOCAL/*
 juju deployer can also include files from the workspace local directory
  ssl_cert: include-base64://{{local_dir}}/$MYCERT.crt


Continuous Integration
======================

contrib/jenkins/mojo-to-jenkins
-------------------------------
The mojo-to-jenkins script uses the contrib/jenkins/job-templates to generate
jenkins jobs. Jenkins can then execute mojo runs against the given
specification. 

cd contrib/jenkins/mojo-to-jenkins
./mojo-to-jenkins $MOJO_PROJECT $SPEC_URL


mojo monitor
------------
Mojo monitor will check the primary collect file of a spec for updated bazaar
branches. The first run of mojo monitor creates state files in the project
directory with the current bzr revision numbers of each branch. Subsequent runs
check against (and update) state files and exits with return code 100 if there
are new revision numbers to test and 0 if not.

Can be used in cron to monitor for new bzr branches and execute jenkins jobs
when changes are found.

mojo monitor --project $MOJO_PROJECT $SPEC_URL monitor


FAQ
===

When do I run mojo with sudo?

  For project creation and destruction use sudo
   sudo mojo project-new $MOJO_PROJECT
  For everything else do not use sudo
   mojo *
  Note: mojo itself will call subcommands with sudo. See 
  contrib/99-mojo-sudoers for details

Why do I get /var/lib/lxc/$MOJO_PROJECT/config permission denied?

  Depending on which version of lxc is installed it may be necessary to
  widen the permissions on /var/lib/lxc. If one sees permission denied on
  /var/lib/lxc/$MOJO_PROJECT/config then chmod the lxc directory:

  chmod 755 /var/lib/lxc

Where should I stage secrets?
  Stage secrets in /srv/mojo/LOCAL/$MOJO_STAGE/
  Add the mojo secrets phase to the manifest file
  This will copy /srv/mojo/LOCAL/$MOJO_STAGE/* $MOJO_WORKSPACE/local
  Where it can be used by juju deployer and other specification scripts.

How can I use shared code for multiple specs?
  The collect phase can be used to collect code and store it in any location
  within the workspace relative to $MOJO_WORKSPACE_DIR/build. By placing shared
  code in the $MOJO_SPEC_DIR it creates a $MOJO_STAGE which can be run with
  mojo $CMD --stage $MOJO_STAGE.
  Example:
  # Collect file
  ../spec/shared $SHARED_BRANCH
  # Command
  mojo run --stage shared 
